/*

Copyright (c) 2005, 2011 John Elliott <jce@seasip.demon.co.uk>

This software is provided 'as-is', without any express or implied warranty. 
In no event will the authors be held liable for any damages arising from the 
use of this software.

Permission is granted to anyone to use this software for any purpose, 
including commercial applications, and to alter it and redistribute it 
freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not 
claim that you wrote the original software. If you use this software in a 
product, an acknowledgment in the product documentation would be appreciated
but is not required.

    2. Altered source versions must be plainly marked as such, and must not 
be misrepresented as being the original software.

    3. This notice may not be removed or altered from any source distribution.

*/
#include <stdio.h>
#include <stdlib.h>
#include <dos.h>

typedef unsigned char byte;

void syntax()
{
	fprintf(stderr, "Syntax: SAVEROM romfile {size-in-k}\n");
	exit(0);
}


int compare(unsigned a, unsigned b, unsigned len)
{
	byte far *p1 = MK_FP(0xF000, a);	
	byte far *p2 = MK_FP(0xF000, b);	

	while (len)
	{
		if (p1[0] != p2[0]) return -1;
		++p1, ++p2, --len;	
	}
	return 0;
}



int guess()
{
	unsigned n;
	volatile byte far *rom = MK_FP(0xF000, 0);

/* Firstly, check for duplication at 32k interval, but not at 16k. */
	if (!compare(0x0000, 0x8000, 0x8000) &&
	     compare(0x8000, 0xC000, 0x4000))
	{
		return 32;
	}
/* Next, check for duplication at 16k interval */
	if (!compare(0x8000, 0xC000, 0x4000))
	{
		return 16;
	}
/* Next, verify if there's ROM there */
	for (n = 0xFFF0; n > 0; n--)
	{
		if (rom[n] != rom[n]) 
		{
			n /= 1024;	/* Size of non-ROM in K */
			return 64 - n;	/* Size of ROM in K */
		}
	}
	return 64;
}


void savepara(FILE *fp, unsigned seg)
{
	unsigned m;
	byte far *rom;
	byte para[16];

	rom   = MK_FP(seg, 0);
	for (m = 0; m < 16; m++)
	{
		para[m] = rom[m];
	}
	if (fwrite(para, 1, sizeof(para), fp) < (int)sizeof(para))
	{
		fprintf(stderr, "Write failed.\n");
		fclose(fp);
		exit(1);
	}
}


void saverom(FILE *fp, unsigned seg, unsigned long paras)
{
	unsigned long n;

	if (paras == 0)	/* Trying to save a full 1M and overflowed the count */
	{
		savepara(fp, 0);
		paras = 65535;
		++seg;
	}
	for (n = 0; n < paras; n++)
	{
		savepara(fp, seg + n);
	}
}

void analyse(int save)
{
	unsigned seg;
	unsigned char far *rom;
	unsigned len, n;
	unsigned char cksum;
	FILE *fp;
	char buf[50];

	for (seg = 0xC000; seg < 0xF000; seg += 2)
	{
		rom = MK_FP(seg, 0);

		if (rom[0] == 0x55 && rom[1] == 0xAA)
		{
			len = 512 * rom[2];
			for (n = cksum = 0; n < len; n++)
			{
				cksum += rom[n];
			}
			if (cksum == 0)
			{
				printf("  ROM signature at segment 0x%04x, length 0x%05x", seg, len);
				fflush(stdout);
				if (save)
				{
					sprintf(buf, "%04x.rom", seg);
					fp = fopen(buf, "wb");
					if (!fp)
					{
						perror(buf);
						exit(1);
					}
					saverom(fp, seg, 32L * rom[2]);
					fclose(fp);
					printf(" saved as %s\n", buf);	
				}
				else putchar('\n');
			}
		}
	}
}


int main(int argc, char **argv)
{
	int romsize; 
	unsigned seg;
	FILE *fp;

	if (argc < 2)
	{
		syntax();
	}
	else if (argc < 3) 
	{
		romsize = guess();
	}
	else
	{
		romsize = atoi(argv[2]);
		if (romsize <= 0 || romsize > 1024)
		{
			syntax();
		}
	}
	fp = fopen(argv[1], "wb");
	if (!fp)
	{
		perror(argv[1]);
		return 1;
	}
	if (argc >= 3)
	{
		analyse(0);
	}
	seg = ((1024 - romsize) * 64);
	printf("%4dK of memory at segment 0x%04x, length 0x%05lx", 
			romsize, seg, 1024L * romsize);
	fflush(stdout);
	saverom(fp, seg, 64L * romsize);
	fclose(fp);
	printf(" saved as %s\n", argv[1]);
	if (argc < 3)
	{
		analyse(1);
	}
	return 0;
}

